from ticketing import conf
from django.test import TestCase
from ticketing.models import get_ticket

import logging
log = logging.getLogger(__name__)

class TicketingTest(TestCase):
    
    def test_get_ticket(self):
        t = get_ticket()
        self.assertTrue(t)
        self.assertTrue(t > 0, "Ticket %s is not greater than zero!" % t)
    
    def test_get_tickets(self):
        t0 = get_ticket()
        self.assertTrue(t0 > 0, "Ticket %s is not greater than zero!" % t0)
        
        t1 = get_ticket()
        self.assertTrue(t1 > 0, "Ticket %s is not greater than zero!" % t1)
        self.assertTrue(t1 > t0, "Ticket %s is not greater than ticket %s!" % (t1, t0))
    
    def test_get_n_tickets(self):
        import random
        n = int((random.random() + 1) * 100)
        
        t0 = get_ticket()
        self.assertTrue(t0 > 0, "Ticket %s is not greater than zero!" % t0)
        
        for i in range(0, n):
            t1 = get_ticket()
            self.assertTrue(t1 > 0, "Ticket %s is not greater than zero!" % t1)
            self.assertTrue(t1 > t0, "Ticket %s is not greater than ticket %s!" % (t1, t0))
            t0 = t1
    
    def test_get_from_sequence(self):
        self.assertTrue(len(conf.SEQUENCES) > 1, "There aren't multiple sequences to test! - %s" % str(conf.SEQUENCES))
        last_seq = 2
        for seq in conf.SEQUENCES:
            t0 = get_ticket(seq)
            self.assertTrue(t0 > 0, "Ticket %s [for sequence %s] is not greater than zero!" % (t0, seq))
            self.assertTrue(t0 < last_seq, "Ticket %s [for sequence %s] is not less than previous sequence %s!" % (t0, seq, last_seq))
            for i in range(0, 10):
                t1 = get_ticket(seq)
                self.assertTrue(t1 > 0, "Ticket %s [for sequence %s] is not greater than zero!" % (t1, seq))
                self.assertTrue(t1 > t0, "Ticket %s [for sequence %s] is not greater than ticket %s!" % (t1, seq, t0))
                t0 = t1
            last_seq = t0
    
    def test_get_sequences_in_order(self):
        self.assertTrue(len(conf.SEQUENCES) > 1, "There aren't multiple sequences to test! - %s" % str(conf.SEQUENCES))
        for i in range(1, 10):
            for seq in conf.SEQUENCES:
                t = get_ticket(seq)
                self.assertTrue(t > 0, "Ticket %s [for sequence %s] is not greater than zero!" % (t, seq))
                self.assertTrue(t == i, "Ticket %s [for sequence %s] is not the expected value of %s!" % (t, seq, i))
    
    def test_bad_sequence(self):
        from ticketing.exceptions import BadSequenceName
        self.assertRaises(BadSequenceName, get_ticket, ('bad_sequence', ))
    





from ticketing import conf
from django.db import models

class TicketingManager(models.Manager):
    
    def create(self):
        id = self.get_ticket()
        return super(TicketingManager, self).get_query_set().get(pk=id)
    
    def get_query_set(self):
        raise NotImplementedError()
    
    def get_empty_query_set(self):
        raise NotImplementedError()
    
    def get_ticket(self):
        from django.db import connections, transaction
        cursor = connections[self.db].cursor()
        
        sql = "REPLACE INTO `%s` " % (self.model._meta.db_table)
        sql += "(`stub`) VALUES (%s)"
        
        result = cursor.execute(sql, [True])
        transaction.commit_unless_managed(using=self.db)
        if hasattr(cursor, 'lastrowid') and cursor.lastrowid:
            return cursor.lastrowid
        else:
            # Hack. Select the row and get the id manually
            object = super(TicketingManager, self).get_query_set()\
                .filter(stub=True)[0]
            return object.id
    



from ticketing import conf
from django.db import models

class BigAutoField(models.AutoField):
    
    def db_type(self, *args, **kwargs):
        connection = kwargs.get('connection', None)
        if connection and connection.settings_dict['ENGINE'] == 'django.db.backends.mysql':
            return 'bigint UNSIGNED AUTO_INCREMENT'
        else:
            return super(BigAutoField, self).db_type(*args, **kwargs)
    

class TicketField(models.BigIntegerField):
    
    description = "A big integer pre-populated from a ticket"
    
    def __init__(self, sequence=None, **kwargs):
        """
        Set up this field with the name of the sequence to use
        """
        if sequence:
            self.sequence = sequence
        else:
            # Use the default sequence
            self.sequence = conf.DEFAULT_SEQUENCE
        kwargs['editable'] = False
        super(TicketField, self).__init__(**kwargs)
    
    def pre_save(self, model_instance, add):
        """
        When saving the DB, if we are adding this instance, get a ticket
        value for this field. Otherwise pull it from the model instance.
        """
        if add:
            from ticketing.models import get_ticket
            value = get_ticket(self.sequence)
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super(TicketField, self).pre_save(model_instance, add)
    
    def db_type(self, *args, **kwargs):
        connection = kwargs.get('connection', None)
        if connection and connection.settings_dict['ENGINE'] == 'django.db.backends.mysql':
            return 'bigint UNSIGNED'
        else:
            return super(BigAutoField, self).db_type(*args, **kwargs)
    
    def south_field_triple(self):
        "Returns a suitable description of this field for South."
        # We'll just introspect the _actual_ field.
        from south.modelsinspector import introspector
        field_class = 'ticketing.fields.TicketField'
        args, kwargs = introspector(self)
        # That's our definition!
        return (field_class, args, kwargs)

try:
    from south.modelsinspector import add_introspection_rules
    add_introspection_rules(rules=[
        (
            [TicketField], # Class(es) these apply to
            [],         # Positional arguments (not used)
            {           # Keyword argument
                "sequence": ["sequence", {"default": None}],
            },
        ),
    ], patterns=["^ticketing\.fields\.TicketField"])
except ImportError:
    pass



from django.conf import settings

# Testing and Debug settings
TESTING = getattr(settings, 'TICKETING_TESTING', False)
DEBUG = getattr(settings, 'DEBUG', False) and not TESTING

# Let the user redefine the APP_LABEL, which is used for the table
# prefixes.
APP_LABEL = getattr(settings, 'TICKETING_APP_LABEL', 'ticketing')

# A list of names that define all the different ticketing tables. Every
# sequence is a sequence of tickets that can be retreived.
SEQUENCES = getattr(settings, 'TICKETING_SEQUENCES', ('default', ))

# When testing, add a second sequence so we can test specifying sequences
if TESTING:
    SEQUENCES = SEQUENCES + ('test', )

# The name of the default sequence that is queried when creating tickets.
DEFAULT_SEQUENCE = getattr(settings, 'TICKETING_DEFAULT_SEQUENCE', 'default')

if DEFAULT_SEQUENCE not in SEQUENCES:
    msg = ("Cannot specify a default sequence '%s' that " +
        "is not in the list of allowed sequences %s") % (
        DEFAULT_SEQUENCE, str(SEQUENCES)
    )
    from ticketing.exceptions import ImproperlyConfigured
    raise ImproperlyConfigured(msg)




from ticketing import conf
from django.db import models
from django.db.models.signals import post_syncdb
from ticketing.fields import BigAutoField
from ticketing.managers import TicketingManager

# Container for all the ticketing models
ticketing_models = {}

class BaseTicketing(models.Model):
    """
    The base class used by all Ticket generator models.
    
    Define the fields, methods and manager used by all
    classes that will inherit from this abstract model.
    All ticketing models that are created for each
    sequence defined in the configuration will inherit
    these properties.
    
    """
    
    # Explicitly set the id of the model, even though it is the same
    # as the one Django gives it.
    id = BigAutoField(null=False, primary_key=True)
    
    # This is just the smallest placeholder we can create that we can
    # replace into to generate a new id.
    stub = models.BooleanField(null=False, default=True, unique=True)
    
    # Override the default manager
    objects = TicketingManager()
    
    @property
    def ticket(self):
        """Alias for id"""
        return self.id
    
    def __unicode__(self):
        return u'Ticket %d' % self.ticket
    
    def __str__(self):
        return str(unicode(self))
    
    class Meta:
        abstract = True

# Create each model for the ones defined in the configuration
for seq in conf.SEQUENCES:
    class Meta:
        pass
    
    # Set the application label
    setattr(Meta, 'app_label', conf.APP_LABEL)
    
    # __module__ is used with app_label to create the name of the table.
    # In addition, we want to have a Meta class so that the Django code
    # will work properly.
    attrs = {
        '__module__': seq,
        'Meta': Meta,
    }
    
    # Create the model definition
    model = type(seq, (BaseTicketing,), attrs)
    
    # Save the definition in a variable that we can use later. This will
    # also let Django find this model with some python dict traversal
    # goodness.
    ticketing_models[seq] = model

def get_ticket(sequence=conf.DEFAULT_SEQUENCE):
    """
    Get a ticket for the specified sequence.
    
    `sequence` defaults to the configuration parameter DEFAULT_SEQUENCE
    
    """
    if sequence not in ticketing_models:
        msg = "Sequence '%s' is not one of the configured sequences %s" % (
            sequence, conf.SEQUENCES
        )
        from ticketing.exceptions import BadSequenceName
        raise BadSequenceName(msg)
    ticket = ticketing_models[sequence].objects.create()
    return ticket.ticket
